#!/bin/sh

echo -e "BASHISM" \
    "something else \n"

#exec ${loclibdir}/tkcon.tcl \
#        -eval "source ${loclibdir}/console.tcl" \
#        -slave "package require Tk; set argc $#; set argv [list $*]; \
#        source ${loclibdir}/xcircuit.tcl"
